#!/usr/bin/env python

from bisect import bisect_left, insort_left
from email import message_from_file
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP
from vCalendar import parse, ParseError, SECTION_TYPES
import imip_store
import sys

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

OWNER = "resource+manager@example.com"

MESSAGE_SUBJECT = "Calendar system message"

MESSAGE_TEXT = """\
This is a response to a calendar message sent by your calendar program.
"""

# Postfix exit codes.

EX_USAGE        = 64
EX_DATAERR      = 65
EX_NOINPUT      = 66
EX_NOUSER       = 67
EX_NOHOST       = 68
EX_UNAVAILABLE  = 69
EX_SOFTWARE     = 70
EX_OSERR        = 71
EX_OSFILE       = 72
EX_CANTCREAT    = 73
EX_IOERR        = 74
EX_TEMPFAIL     = 75
EX_PROTOCOL     = 76
EX_NOPERM       = 77
EX_CONFIG       = 78

# Permitted iTIP content types.

itip_content_types = [
    "text/calendar",                        # from RFC 6047
    "text/x-vcalendar", "application/ics",  # other possibilities
    ]

# Content interpretation.

def get_itip_structure(elements):
    d = {}
    for name, attr, value in elements:
        if not d.has_key(name):
            d[name] = []
        if name in SECTION_TYPES:
            d[name].append((get_itip_structure(value), attr))
        else:
            d[name].append((value, attr))
    return d

def get_structure_items(d):
    items = []
    for name, value in d.items():
        if isinstance(value, list):
            for v, a in value:
                items.append((name, a, v))
        else:
            v, a = value
            items.append((name, a, get_structure_items(v)))
    return items

def get_items(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0]
        else:
            return values
    else:
        return None

def get_item(d, name):
    return get_items(d, name, False)

def get_value_map(d, name):
    items = get_items(d, name)
    if items:
        return dict(items)
    else:
        return {}

def get_values(d, name, all=True):
    if d.has_key(name):
        values = d[name]
        if not all and len(values) == 1:
            return values[0][0]
        else:
            return map(lambda x: x[0], values)
    else:
        return None

def get_value(d, name):
    return get_values(d, name, False)

def get_address(value):
    return value.startswith("mailto:") and value[7:] or value

def get_uri(value):
    return value.startswith("mailto:") and value or "mailto:%s" % value

# Time management.

def insert_period(freebusy, period):
    insort_left(freebusy, period)

def remove_period(freebusy, uid):
    i = 0
    while i < len(freebusy):
        t = freebusy[i]
        if len(t) >= 3 and t[2] == uid:
            del freebusy[i]
        else:
            i += 1

def period_overlaps(freebusy, period):
    dtstart, dtend = period[:2]
    i = bisect_left(freebusy, (dtstart, dtend, None))
    return (
        i < len(freebusy) and (dtend is None or freebusy[i][0] < dtend)
        or
        i > 0 and freebusy[i - 1][1] > dtstart
        )

# Sending of outgoing messages.

def sendmail(sender, recipients, data):
    smtp = SMTP("localhost")
    smtp.sendmail(sender, recipients, data)
    smtp.quit()

# Processing of incoming messages.

def process(f, original_recipients, recipients):

    """
    Process content from the stream 'f' accompanied by the given
    'original_recipients' and 'recipients'.
    """

    msg = message_from_file(f)
    senders = msg.get_all("Reply-To") or msg.get_all("From")

    # Handle messages with iTIP parts.

    all_parts = []

    for part in msg.walk():
        if part.get_content_type() in itip_content_types and \
            part.get_param("method"):

            all_parts += handle_itip_part(part, original_recipients)

    # Pack the parts into a single message.

    if all_parts:
        text_part = MIMEText(MESSAGE_TEXT)
        all_parts.insert(0, text_part)
        message = MIMEMultipart("alternative", _subparts=all_parts)
        message.preamble = MESSAGE_TEXT

        message["From"] = OWNER
        for sender in senders:
            message["To"] = sender
        message["Subject"] = MESSAGE_SUBJECT

        if "-d" in sys.argv:
            print message
        else:
            sendmail(OWNER, senders, message.as_string())

def to_part(method, calendar):

    """
    Write using the given 'method', the 'calendar' details to a MIME
    text/calendar part.
    """

    encoding = "utf-8"
    out = StringIO()
    try:
        calendar[:0] = [
            ("METHOD", {}, method),
            ("VERSION", {}, "2.0")
            ]
        imip_store.to_stream(out, calendar, "VCALENDAR", encoding)
        part = MIMEText(out.getvalue(), "calendar", encoding)
        part.set_param("method", method)
        return part

    finally:
        out.close()

def parse_object(f, encoding, objtype):

    """
    Parse the iTIP content from 'f' having the given 'encoding'. Return None if
    the content was not readable or suitable.
    """

    try:
        try:
            doctype, attrs, elements = parse(f, encoding=encoding)
            if doctype == objtype:
                return get_itip_structure(elements)
        finally:
            f.close()
    except (ParseError, ValueError):
        pass

    return None

def handle_itip_part(part, recipients):

    "Handle the given iTIP 'part' for the given 'recipients'."

    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    itip = parse_object(f, part.get_content_charset(), "VCALENDAR")
    if not itip:
        sys.exit(EX_DATAERR)

    # Only handle calendar information.

    all_parts = []

    # Require consistency between declared and employed methods.

    if get_value(itip, "METHOD") == method:

        # Look for different kinds of sections.

        all_objects = []

        for name, cls in handlers:
            for details in get_values(itip, name) or []:

                # Dispatch to a handler and obtain any response.

                handler = cls(details, recipients)
                object = methods[method](handler)()

                # Concatenate responses for a single calendar object.

                if object:
                    all_objects += object

        # Obtain a message part for the objects.

        if all_objects:
            all_parts.append(to_part(response_methods[method], all_objects))

    return all_parts

class Handler:

    "General handler support."

    def __init__(self, details, recipients):

        """
        Initialise the handler with the 'details' of a calendar object and the
        'recipients' of the object.
        """

        self.details = details
        self.recipients = set(recipients)

        self.uid = get_value(details, "UID")
        self.sequence = get_value(details, "SEQUENCE")
        self.dtstamp = get_value(details, "DTSTAMP")

        self.store = imip_store.FileStore()

    def get_items(self, name, all=True):
        return get_items(self.details, name, all)

    def get_item(self, name):
        return get_item(self.details, name)

    def get_value_map(self, name):
        return get_value_map(self.details, name)

    def get_values(self, name, all=True):
        return get_values(self.details, name, all)

    def get_value(self, name):
        return get_value(self.details, name)

    def filter_by_recipients(self, values):
        return self.recipients.intersection(map(get_address, values))

    def require_organiser_and_attendees(self):
        attendee_map = self.get_value_map("ATTENDEE")
        organiser = self.get_item("ORGANIZER")

        # Only provide details for recipients who are also attendees.

        attendees = {}
        for attendee in map(get_uri, self.filter_by_recipients(attendee_map)):
            attendees[attendee] = attendee_map[attendee]

        if not attendees and not organiser:
            return None

        return organiser, attendees

class Event(Handler):

    "An event handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def declinecounter(self):

        """
        Since this handler does not send counter proposals, it will not handle
        replies to such proposals.
        """

        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):

        """
        Respond to a request by preparing a reply containing accept/decline
        information for each indicated attendee.

        No support for countering requests is implemented.
        """

        oa = self.require_organiser_and_attendees()
        if not oa:
            return None

        (organiser, organiser_attr), attendees = oa

        # Process each attendee separately.

        for attendee, attendee_attr in attendees.items():

            # Check for event using UID.

            f = self.store.get_event(attendee, self.uid)
            event = f and parse_object(f, "utf-8", "VEVENT")

            # If found, compare SEQUENCE and potentially DTSTAMP.

            if event:
                sequence = get_value(event, "SEQUENCE")
                dtstamp = get_value(event, "DTSTAMP")

                # If the request refers to an older version of the event, ignore
                # it.

                old_dtstamp = self.dtstamp < dtstamp

                if sequence is not None and (
                    int(self.sequence) < int(sequence) or
                    int(self.sequence) == int(sequence) and old_dtstamp
                    ) or old_dtstamp:

                    continue

            # If newer than any old version, discard old details from the
            # free/busy record and check for suitability.

            dtstart = self.get_value("DTSTART")
            dtend = self.get_value("DTEND")

            conflict = False
            freebusy = self.store.get_freebusy(attendee)

            if freebusy:
                remove_period(freebusy, self.uid)
                conflict = period_overlaps(freebusy, (dtstart, dtend))
            else:
                freebusy = []

            # If the event can be scheduled, it is registered and a reply sent
            # accepting the event. (The attendee has PARTSTAT=ACCEPTED as an
            # attribute.)

            if not conflict:
                insert_period(freebusy, (dtstart, dtend, self.uid))
                self.store.set_freebusy(attendee, freebusy)
                self.store.set_event(attendee, self.uid, get_structure_items(self.details))
                attendee_attr["PARTSTAT"] = "ACCEPTED"

            # If the event cannot be scheduled, it is not registered and a reply
            # sent declining the event. (The attendee has PARTSTAT=DECLINED as an
            # attribute.)

            else:
                attendee_attr["PARTSTAT"] = "DECLINED"

            self.details["ATTENDEE"] = [(attendee, attendee_attr)]
            return [("VEVENT", {}, get_structure_items(self.details))]

class Freebusy(Handler):

    "A free/busy handler."

    def publish(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):

        """
        Respond to a request by preparing a reply containing free/busy
        information for each indicated attendee.
        """

        oa = self.require_organiser_and_attendees()
        if not oa:
            return None

        (organiser, organiser_attr), attendees = oa

        # Construct an appropriate fragment.

        calendar = []
        cwrite = calendar.append

        # Get the details for each attendee.

        for attendee, attendee_attr in attendees.items():
            freebusy = self.store.get_freebusy(attendee)

            if freebusy:
                record = []
                rwrite = record.append

                rwrite(("ORGANIZER", organiser_attr, organiser))
                rwrite(("ATTENDEE", attendee_attr, attendee))
                rwrite(("UID", {}, self.uid))

                for start, end, uid in freebusy:
                    rwrite(("FREEBUSY", {}, [start, end]))

                cwrite(("VFREEBUSY", {}, record))

        # Return the reply.

        return calendar

class Journal(Handler):

    "A journal entry handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def publish(self):
        pass

class Todo(Handler):

    "A to-do item handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def declinecounter(self):

        """
        Since this handler does not send counter proposals, it will not handle
        replies to such proposals.
        """

        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):
        pass

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

response_methods = {
    "REQUEST"        : "REPLY",
    }

def main():

    # Obtain the different kinds of recipients.

    original_recipients = []
    recipients = []

    l = []

    for arg in sys.argv[1:]:
        if arg == "-o":
            l = original_recipients
        elif arg == "-r":
            l = recipients
        elif arg == "-d":
            pass
        else:
            l.append(arg)

    process(sys.stdin, original_recipients, recipients)

if __name__ == "__main__":
    if "-d" in sys.argv[1:]:
        main()
    else:
        try:
            main()
        except SystemExit, value:
            sys.exit(value)
        except Exception, exc:
            type, value, tb = sys.exc_info()
            print >>sys.stderr, "Exception %s at %d" % (exc, tb.tb_lineno)
            sys.exit(EX_TEMPFAIL)
    sys.exit(0)

# vim: tabstop=4 expandtab shiftwidth=4
